# ovirt-imageio
# Copyright (C) 2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
"""
_internal - imageio internal moudules used by the server and client.

Do not use modules form this package; they are internal implemention details
and may change on each release in incompatible way.
"""
